"""
.. module:: sink
   :synopsis: Nuts that reads streams and potentially produce non-stream output.
"""
from __future__ import absolute_import

import sys
import csv
import math
import six

import collections as cl

from six.moves import reduce, zip, range
from nutsflow.base import NutSink
from nutsflow.factory import nut_sink
from nutsflow.common import as_tuple, is_iterable, colfunc
from nutsflow.iterfunction import nth, consume, length, take


@nut_sink
def Sort(iterable, key=None, reverse=False):
    """
    iterable >> Sort(key=None, reverse=False)

    Sorts iterable with respect to key function or column index(es).

    >>> [3, 1, 2] >> Sort()
    [1, 2, 3]

    >>> [3, 1, 2] >> Sort(reverse=True)
    [3, 2, 1]

    >>> [(1,'c'), (2,'b'), (3,'a')] >> Sort(1)
    [(3, 'a'), (2, 'b'), (1, 'c')]

    >>> ['a3', 'c1', 'b2'] >> Sort(key=lambda s: s[0])
    ['a3', 'b2', 'c1']

    >>> ['a3', 'c1', 'b2'] >> Sort(key=0)
    ['a3', 'b2', 'c1']

    >>> ['a3', 'c1', 'b2'] >> Sort(1)
    ['c1', 'b2', 'a3']

    >>> ['a3', 'c1', 'b2'] >> Sort((1,0))
    ['c1', 'b2', 'a3']

    :param iterable iterable: Iterable
    :param int|tuple|function|None key: function to sort based on or
           column index(es) tuples/vectors/strings are sorted by.
    :param boolean reverse: True: reverse order.
    :return: Sorted iterable
    :rtype: list
    """
    return sorted(iterable, key=colfunc(key), reverse=reverse)


@nut_sink
def Sum(iterable, key=None):
    """
    iterable >> Sum(key=None)

    Return sum over inputs (transformed or extracted by key function)

    >>> [1, 2, 3] >> Sum()
    6

    >>> [1, 2, 3] >> Sum(lambda x: x*x)
    14

    >>> data = [(1, 10), (2, 20), (3, 30)]
    >>> data >> Sum(key=0)
    6
    >>> data >> Sum(key=1)
    60

    :param iterable iterable: Iterable over numbers
    :param int|tuple|function|None key: Key function to extract elements.
    :return: Sum of numbers
    :rtype: number
    """
    return sum(map(colfunc(key), iterable))


@nut_sink
def Mean(iterable, key=None, default=None):
    """
    iterable >> Mean(key=None, default=None)

    Return mean value of inputs (transformed or extracted by key function).

    >>> [1, 2, 3] >> Mean()
    2.0

    >>> [] >> Mean(default=0)
    0

    >>> data = [(1, 10), (2, 20), (3, 30)]
    >>> data >> Mean(key=0)
    2.0
    >>> data >> Mean(key=1)
    20.0

    :param iterable iterable: Iterable over numbers
    :param object default: Value returned if iterable is empty.
    :param int|tuple|function|None key: Key function to extract elements.
    :return: Mean of numbers or default value
    :rtype: number
    """
    f = colfunc(key)
    sum, n = 0, 0
    for e in iterable:
        sum += f(e)
        n += 1
    return float(sum) / n if n else default


@nut_sink
def MeanStd(iterable, key=None, default=None, ddof=1):
    """
    iterable >> MeanStd(key=None, default=None, ddof=1)

    Return mean and standard deviation of inputs (transformed or extracted
    by key function).
    Standard deviation is with degrees of freedom = 1

    >>> [1, 2, 3] >> MeanStd()
    (2.0, 1.0)

    >>> data = [(1, 10), (2, 20), (3, 30)]
    >>> data >> MeanStd(key=0)
    (2.0, 1.0)
    >>> data >> MeanStd(1)
    (20.0, 10.0)

    :param iterable iterable: Iterable over numbers
    :param object default: Value returned if iterable is empty.
    :param int|tuple|function|None key: Key function to extract elements.
    :param int ddof: Delta degrees of freedom (should 0 or 1)
    :return: Mean and standard deviation of numbers or default value
    :rtype: tuple (mean, std)
    """
    f = colfunc(key)
    sume, sqre, n = 0.0, 0.0, 0.0
    for e in iterable:
        e = f(e)
        sume += e
        sqre += e * e
        n += 1.0
    if n - ddof <= 0:
        return default
    avg = sume / n
    dev = math.sqrt((n * sqre - sume * sume) / (n * (n - ddof)))
    return avg, dev


@nut_sink
def Max(iterable, key=None, default=None):
    """
    iterable >> Max(key=None, default=None)

    Return maximum of inputs (transformed or extracted by key function).

    >>> [1, 2, 3, 2] >> Max()
    3

    >>> ['1', '123', '12'] >> Max(key=len)
    '123'

    >>> [] >> Max(default=0)
    0

    >>> data = [(3, 10), (2, 20), (1, 30)]
    >>> data >> Max(key=0)
    (3, 10)

    >>> data >> Max(1)
    (1, 30)


    :param iterable iterable: Iterable over numbers
    :param int|tuple|function|None key: Key function to extract or
           transform elements. None = identity function.
    :param object default: Value returned if iterable is empty.
    :return: largest element according to key function
    :rtype: object
    """
    try:
        return max(iterable, key=colfunc(key))
    except Exception:
        return default


@nut_sink
def Min(iterable, key=None, default=None):
    """
    iterable >> Min(key=None, default=None)

    Return minimum of inputs (transformed or extracted by key function).

    >>> [1, 2, 3, 2] >> Min()
    1

    >>> ['1', '123', '12'] >> Min(key=len)
    '1'

    >>> [] >> Min(default=0)
    0

    >>> data = [(3, 10), (2, 20), (1, 30)]
    >>> data >> Min(key=0)
    (1, 30)

    >>> data >> Min(1)
    (3, 10)


    :param iterable iterable: Iterable over numbers
    :param int|tuple|function|None key: Key function to extract or
           transform elements. None = identity function.
    :param object default: Value returned if iterable is empty.
    :return: smallest element according to key function
    :rtype: object
    """
    try:
        return min(iterable, key=colfunc(key))
    except Exception:
        return default


@nut_sink
def ArgMax(iterable, key=None, default=None, retvalue=False):
    """
    iterable >> ArgMax(key=None, default=None, retvalue=False)

    Return index of first maximum element (and maximum) in input
    (transformed or extracted by key function).

    >>> [1, 2, 0, 2] >> ArgMax()
    1

    >>> ['12', '1', '123'] >> ArgMax(key=len, retvalue=True)
    (2, '123')

    >>> ['12', '1', '123'] >> ArgMax(key=len)
    2

    >>> [] >> ArgMax(default=0)
    0

    >>> [] >> ArgMax(default=(None, 0), retvalue=True)
    (None, 0)

    >>> data = [(3, 10), (2, 20), (1, 30)]
    >>> data >> ArgMax(key=0)
    0
    >>> data >> ArgMax(1)
    2

    :param iterable iterable: Iterable over numbers
    :param int|tuple|function|None key: Key function to extract or
           transform elements. None = identity function.
    :param object default: Value returned if iterable is empty.
    :param bool retvalue: If True the index and the value of the
           maximum element is returned.
    :return: index of largest element according to key function
             and the largest element itself if retvalue==True
    :rtype: object | tuple
    """
    try:
        f = colfunc(key)
        i, v = max(enumerate(iterable), key=lambda i_e: f(i_e[1]))
        return (i, v) if retvalue else i
    except Exception:
        return default


@nut_sink
def ArgMin(iterable, key=None, default=None, retvalue=False):
    """
    iterable >> ArgMin(key=None, default=None, retvalue=True)

    Return index of first minimum element (and minimum) in input
    (transformed or extracted by key function).

    >>> [1, 2, 0, 2] >> ArgMin()
    2

    >>> ['12', '1', '123'] >> ArgMin(key=len, retvalue=True)
    (1, '1')

    >>> ['12', '1', '123'] >> ArgMin(key=len)
    1

    >>> [] >> ArgMin(default=0)
    0

    >>> [] >> ArgMin(default=(None, 0), retvalue=True)
    (None, 0)

    >>> data = [(3, 10), (2, 20), (1, 30)]
    >>> data >> ArgMin(key=0)
    2
    >>> data >> ArgMin(1)
    0


    :param iterable iterable: Iterable over numbers
    :param int|tuple|function|None key: Key function to extract or
           transform elements. None = identity function.
    :param object default: Value returned if iterable is empty.
    :param bool retvalue: If True the index and the value of the
           minimum element is returned.
    :return: index of smallest element according to key function
             and the smallest element itself if retvalue==True.
    :rtype: object | tuple
    """
    try:
        f = colfunc(key)
        i, v = min(enumerate(iterable), key=lambda i_e1: f(i_e1[1]))
        return (i, v) if retvalue else i
    except Exception:
        return default


Reduce = nut_sink(reduce, 1)
"""
iterable >> Reduce(func [,initiaizer])

Reduces the iterable using the given function.
See https://docs.python.org/2/library/functions.html#reduce

>>> [1, 2, 3] >> Reduce(lambda a,b: a+b)
6

>>> [2] >> Reduce(lambda a,b: a*b, 1)
2

:param iterable iterable: Any iterable
:param function func: Reduction function
:return: Result of reduction
:rtype: any
"""

Nth = nut_sink(nth)
"""
iterable >> Nth(nth)

Return n-th element of iterable. This consumes the iterable!

>>> 'test' >> Nth(2)
s

:param iterable iterable: Any iterable
:param int nth: Index of element in iterable to return
:return: n-th element
:rtype: any
"""

Next = nut_sink(next)
"""
iterable >> Next()

Return next element of iterable.

>>> [1,2,3] >> Next()
1

:param iterable iterable: Any iterable
:return: next element
:rtype: any
"""

Consume = nut_sink(consume)
"""
iterable >> Consume(n=None)

Consume n elements of the iterable.

>>> [1,2,3] >> Print() >> Consume()   # Without Consume nothing happens!
1
2
3

>>> [1,2,3] >> Print() >> Consume(2)
1
2

:param iterable iterable: Iterable
:param int n: Number of elements to consume.
              n = None means the whole iterable is consumed.

"""

Count = nut_sink(length)
"""
iterable >> Count()

Return number elements in input iterable.  This consumes the iterable!

>>> [0, 1, 2] >> Count()
3

:param iterable iterable: Any iterable
:return: Number elements in interable
:rtype: int
"""


@nut_sink
def Unzip(iterable, container=None):
    """
    iterable >> Unzip(container=None)

    Same as izip(*iterable) but returns iterators for container=None

    >>> [(1, 2, 3), (4, 5, 6)] >> Unzip(tuple) >> Collect()
    [(1, 4), (2, 5), (3, 6)]

    :param iterable iterable:  Any iterable, e.g. list, range, ...
    :param container container: If not none, unzipped results are collected
       in the provided container, eg. list, tuple, set
    :return: Unzip iterable.
    :rtype: iterator over iterators
    """
    unzipped = zip(*iterable)
    return map(container, unzipped) if container else unzipped


@nut_sink
def Head(iterable, n, container=list):
    """
    iterable >> Head(n, container=list)

    Collect first n elements of iterable in specified container.

    >>> [1, 2, 3, 4] >> Head(2)
    [1, 2]

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param int n: Number of elements to take.
    :param container container: Container to collect elements in, e.g. list, set
    :return: Container with head elements
    :rtype: container
    """
    return container(take(iterable, n))


@nut_sink
def Tail(iterable, n, container=list):
    """
    iterable >> Tail(n, container=list)

    Collect last n elements of iterable in specified container. This consumes
    the iterable completely!

    >>> [1, 2, 3, 4] >> Tail(2)
    [3, 4]

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param int n: Number of elements to take.
    :param container container: Container to collect elements in, e.g. list, set
    :return: Container with tail elements
    :rtype: container
    """
    return container(cl.deque(iterable, n))


@nut_sink
def CountValues(iterable, column=None, relative=False):
    """
    iterable >> CountValues(relative=False)

    Return dictionary with (relative) counts of the values
    in the input iterable.

    >>> 'abaacc' >> CountValues()  # doctest: +SKIP
    {'a': 3, 'b': 1, 'c': 2}

    >>> 'aabaab' >> CountValues(relative=True)  # doctest: +SKIP
    {'a': 1.0, 'b': 0.5}

    >>> data = [('a', 'X'), ('b', 'Y'), ('a', 'Y')]
    >>> data >> CountValues(column=0)  # doctest: +SKIP
    {'a': 2, 'b': 1}
    >>> data >> CountValues(column=1)  # doctest: +SKIP
    {'Y': 2, 'X': 1}

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param int|None column: Column of values in iterable to extract values from.
       If colum=None the values in the iterable themselves will be counted.
    :param bool relative: True: return relative counts otherwise absolute counts
    :return: Dictionary with (relative) counts for elements in iterable.
    :rtype: dict
    """
    values = iterable if column is None else (i[column] for i in iterable)
    cnts = dict(cl.Counter(values))
    if not relative or not cnts.values():
        return cnts
    max_cnt = max(cnts.values())
    n = float(max_cnt) if max_cnt else 1.0
    return {k: v / n for k, v in six.iteritems(cnts)}


@nut_sink
def Collect(iterable, container=list):
    """
    iterable >> Collect(container)

    Collects all elements of the iterable input in the given container.

    >>> range(5) >> Collect()
    [0, 1, 2, 3, 4]

    >>> [1, 2, 3, 2] >> Collect(set)  # doctest: +SKIP
    {1, 2, 3}

    >>> [('one', 1), ('two', 2)] >> Collect(dict)  # doctest: +SKIP
    {'one': 1, 'two': 2}

    :param iterable iterable: Any iterable, e.g. list, range, ...
    :param container container: Some container, e.g. list, set, dict
           that can be filled from an iterable
    :return: Container
    :rtype: container
    """
    return container(iterable)


@nut_sink
def Join(iterable, separator=''):
    """
    iterable >> Join(separator='')

    Same as Python's sep.join(iterable). Concatenates the elements in
    the iterable to a string using the given separator. In addition to
    Python's sep.join(iterable) it also automatically converts elements
    to strings.

    :param iterable iterable: Any iterable
    :param string separator: Seperator string between elements.
    :return: String of with concatenated elements of iterable.
    :rtype: str
    """
    return separator.join(map(str, iterable))


class WriteCSV(NutSink):
    """
    Write data to a CSV file using Python's CSV writer.
    See: https://docs.python.org/2/library/csv.html
    """

    def __init__(self, filepath, cols=None, skipheader=0, flush=False,
                 encoding=None, fmtfunc=lambda x: x, **kwargs):
        """
        WriteCSV(filepath, cols, skipheader, flush, fmtfunc, **kwargs)

        Write data in Comma Separated Values format (CSV) and other formats
        to file. Tab Separated Values (TSV) files can be written by
        specifying a different delimiter. Note that in the docstring below
        delimiter is '\\t' but in code it should be '\t'. See unit tests.

        Also see https://docs.python.org/2/library/csv.html
        and ReadCSV.


        >>> import os
        >>> filepath = 'tests/data/temp_out.csv'
        >>> with WriteCSV(filepath) as writer:
        ...     range(10) >> writer
        >>> os.remove(filepath)

        >>> with WriteCSV(filepath, cols=(1,0)) as writer:
        ...     [(1,2), (3,4)] >> writer
        >>> os.remove(filepath)

        >>> filepath = 'tests/data/temp_out.tsv'
        >>> with WriteCSV(filepath, delimiter='\\t') as writer:
        ...     [[1,2], [3,4]] >> writer
        >>> os.remove(filepath)


        :param string filepath: Path to file in CSV format.
        :param tuple cols: Indices of the columns to write.
                           If None all columns are written.
        :param int skipheader: Number of header rows to skip.
        :param bool flush: If True flush after every line written.
        :param str encoding: Character encoding, e.g. "utf-8"
                             Ignored for Python 2.x!
        :param function fmtfunc: Function to apply to the elements of each row.
        :param kwargs kwargs: Keyword arguments for Python's CSV writer.
                              See https://docs.python.org/2/library/csv.html
        """
        open2 = lambda fp: open(fp, 'w')
        open3 = lambda fp: open(fp, 'w', encoding=encoding)
        is_py3 = sys.version_info >= (3, 0)
        self.csvfile = open3(filepath) if is_py3 else open2(filepath)
        self.columns = cols if cols is None else as_tuple(cols)
        self.flush = flush
        self.fmtfunc = fmtfunc
        self.skipheader = skipheader
        self.writer = csv.writer(self.csvfile, lineterminator='\n', **kwargs)

    def close(self):
        """Close writer"""
        self.csvfile.close()
        self.writer = None

    def __enter__(self):
        """Implementation of context manager API"""
        return self

    def __exit__(self, *args):
        """Implementation of context manager API"""
        self.close()

    def __rrshift__(self, iterable):
        """Write elements of iterable to file"""
        cols = self.columns
        iterable = iter(iterable)
        for _ in range(self.skipheader):
            next(iterable)
        for row in iterable:
            row = row if is_iterable(row) else [row]
            row = [row[i] for i in cols] if cols else row
            row = [self.fmtfunc(r) for r in row]
            self.writer.writerow(row)
            if self.flush:
                self.csvfile.flush()



"""
.. module:: test_common
   :synopsis: Unit tests for common module
"""

from __future__ import print_function

import sys
import time

from pytest import approx
from time import sleep
from nutsflow import MeanStd
from nutsflow.common import (sec_to_hms, timestr, Redirect, as_tuple, as_set,
                             as_list, is_iterable, colfunc, console,
                             itemize, StableRandom, Timer)


def test_is_iterable():
    assert is_iterable([1, 2])
    assert not is_iterable('12')


def test_as_tuple():
    assert as_tuple(1) == (1,)
    assert as_tuple((1, 2)) == (1, 2)
    assert as_tuple([1, 2]) == (1, 2)


def test_as_list():
    assert as_list(1) == [1]
    assert as_list((1, 2)) == [1, 2]
    assert as_list([1, 2]) == [1, 2]


def test_as_set():
    assert as_set(1) == (1,)
    assert as_set((1, 2)) == {1, 2}
    assert as_set([1, 2]) == {1, 2}


def test_itemize():
    assert itemize([]) == []
    assert itemize([3]) == 3
    assert itemize([3, 2, 1]) == [3, 2, 1]


def test_sec_to_hms():
    assert sec_to_hms('80') == (0, 1, 20)
    assert sec_to_hms(3 * 60 * 60 + 2 * 60 + 1) == (3, 2, 1)


def test_timestr():
    assert timestr('') == ''
    assert timestr('80') == '0:01:20'


def test_colfunc():
    data = ['a3', 'b2', 'c1']
    assert list(map(colfunc(None), data)) == data
    assert list(map(colfunc(0), data)) == ['a', 'b', 'c']
    assert list(map(colfunc(1), data)) == ['3', '2', '1']
    expected = [['3', 'a'], ['2', 'b'], ['1', 'c']]
    assert list(map(colfunc((1, 0)), data)) == expected


def test_console():
    with Redirect() as out:
        console('test')
    assert out.getvalue() == 'test\n'


def test_Redirect():
    with Redirect() as out:
        print('test')
    assert out.getvalue() == 'test\n'

    with Redirect('STDERR') as out:
        print('error', file=sys.stderr)
    assert out.getvalue() == 'error\n'


def test_StableRandom():
    rnd = StableRandom(1)
    assert rnd.randint(1, 10) == 5
    assert rnd.uniform(-10, 10) == approx(9.943696167306904)
    assert rnd.random() == approx(0.7203244894557457)
    assert rnd.sample(range(10), 3) == [9, 0, 1]

    lst = [1, 2, 3, 4, 5]
    rnd.shuffle(lst)
    assert lst == [5, 3, 1, 4, 2]

    rnd = StableRandom()
    assert max(rnd.random() for _ in range(1000)) < 1.0
    assert min(rnd.random() for _ in range(1000)) >= 0.0

    rnd1, rnd2 = StableRandom(0), StableRandom(0)
    for _ in range(100):
        assert rnd1.random() == rnd2.random()

    rnd1, rnd2 = StableRandom(0), StableRandom(0)
    rnd2.jumpahead(10)
    for _ in range(100):
        assert rnd1.random() != rnd2.random()
    rnd2.setstate(rnd1.getstate())
    for _ in range(100):
        assert rnd1.random() == rnd2.random()

    rnd1, rnd2 = StableRandom(0), StableRandom(1)
    for _ in range(100):
        assert rnd1.random() != rnd2.random()

    rnd1 = StableRandom()
    sleep(0.5)  # seed is based on system time.
    rnd2 = StableRandom()
    for _ in range(100):
        assert rnd1.random() != rnd2.random()

    rnd = StableRandom()
    numbers = [rnd._randbelow(10) for _ in range(1000)]
    assert max(numbers) < 10
    assert min(numbers) >= 0

    rnd = StableRandom()
    numbers = [rnd.gauss_next() for _ in range(10000)]
    my, std = numbers >> MeanStd()
    assert 0.0 == approx(my, abs=0.1)
    assert 1.0 == approx(std, abs=0.1)


def test_timer():
    t = Timer()
    time.sleep(1.3)
    assert str(t) == '00:01'

    with Timer() as t:
        time.sleep(1.3)
    assert str(t) == '00:01'

    t = Timer()
    time.sleep(0.5)
    t.start()
    time.sleep(1.3)
    t.stop()
    time.sleep(0.5)
    assert str(t) == '00:01'



"""
.. module:: test_sink
   :synopsis: Unit tests for sink module
"""

import os
import pytest

from six.moves import range
from nutsflow import *


def assert_equal_text(text1, text2):
    assert text1.splitlines() == text2.splitlines()


def test_Sort():
    assert [] >> Sort() == []
    assert [3, 1, 2] >> Sort() == [1, 2, 3]
    assert [3, 1, 2] >> Sort(reverse=True) == [3, 2, 1]
    assert ['a3', 'c1', 'b2'] >> Sort(key=0) == ['a3', 'b2', 'c1']
    assert ['a3', 'c1', 'b2'] >> Sort(key=1) == ['c1', 'b2', 'a3']


def test_Sum():
    assert [] >> Sum() == 0

    data = [1, 2, 3]
    assert data >> Sum() == 6
    assert data >> Sum(key=lambda x: x * x) == 14

    data = [(1, 10), (2, 20), (3, 30)]
    assert data >> Sum(key=0) == 6
    assert data >> Sum(key=1) == 60
    assert data >> Sum(key=lambda x: x[0] + x[1]) == 66


def test_Mean():
    assert [] >> Mean() is None
    assert [] >> Mean(default=0) == 0
    assert [1, 2, 3] >> Mean() == pytest.approx(2.0, rel=1e-2)

    data = [(1, 10), (2, 20), (3, 30)]
    assert data >> Mean(key=0) == pytest.approx(2.0, rel=1e-2)
    assert data >> Mean(key=1) == pytest.approx(20.0, rel=1e-2)


def test_MeanStd():
    assert [] >> MeanStd() is None
    assert [] >> MeanStd(ddof=1) is None
    assert [] >> MeanStd(default=0) == 0

    assert [1, 2, 3] >> MeanStd() == pytest.approx([2.0, 1.0], rel=1e-2)
    assert [1, 2, 3] >> MeanStd(ddof=0) == pytest.approx([2.0, 0.81], rel=1e-2)

    data = [(1, 10), (2, 20), (3, 30)]
    assert data >> MeanStd(key=0) == pytest.approx([2.0, 1.0], rel=1e-2)
    assert data >> MeanStd(key=1) == pytest.approx([20.0, 10.0], rel=1e-2)


def test_Max():
    assert [] >> Max(default=0) == 0
    assert [0, 3, 2] >> Max() == 3
    assert ['1', '123', '12'] >> Max(key=len) == '123'

    data = [(3, 10), (2, 20), (1, 30)]
    assert data >> Max(key=0) == (3, 10)
    assert data >> Max(key=1) == (1, 30)


def test_Min():
    assert [] >> Min(default=0) == 0
    assert [3, 1, 2] >> Min() == 1
    assert ['123', '1', '12'] >> Min(key=len) == '1'

    data = [(3, 10), (2, 20), (1, 30)]
    assert data >> Min(key=0) == (1, 30)
    assert data >> Min(key=1) == (3, 10)


def test_ArgMax():
    assert [] >> ArgMax(default=0) == 0
    assert [] >> ArgMax(default=(None, 0)) == (None, 0)

    assert [0, 3, 2] >> ArgMax() == 1

    assert ['1', '123', '12'] >> ArgMax(key=len) == 1
    assert ['1', '123', '12'] >> ArgMax(key=len, retvalue=True) == (1, '123')

    data = [(3, 10), (2, 20), (1, 30)]
    assert data >> ArgMax(key=0) == 0
    assert data >> ArgMax(key=1) == 2


def test_ArgMin():
    assert [] >> ArgMin(default=0) == 0
    assert [] >> ArgMin(default=(None, 0)) == (None, 0)

    assert [3, 1, 2] >> ArgMin() == 1

    assert ['123', '1', '12'] >> ArgMin(key=len) == 1
    assert ['123', '1', '12'] >> ArgMin(key=len, retvalue=True) == (1, '1')

    data = [(3, 10), (2, 20), (1, 30)]
    assert data >> ArgMin(key=0) == 2
    assert data >> ArgMin(key=1) == 0


def test_Reduce():
    assert [] >> Reduce(lambda a, b: a + b, None) == None
    assert [0, 1, 2] >> Reduce(lambda a, b: a + b) == 3
    assert [2] >> Reduce(lambda a, b: a * b, 1) == 2


def test_Nth():
    assert [0, 1, 2, 3] >> Nth(2) == 2


def test_Next():
    it = iter([1, 2, 3])
    assert it >> Next() == 1
    assert it >> Next() == 2
    assert it >> Next() == 3


def test_Consume():
    it = iter([0, 1, 2])
    it >> Consume()
    assert next(it, None) is None


def test_Count():
    assert [] >> Count() == 0
    assert [0, 1, 2, 3] >> Count() == 4


def test_Unzip():
    expected = [(1, 4), (2, 5), (3, 6)]
    input = [(1, 2, 3), (4, 5, 6)]

    assert [] >> Unzip() >> Map(tuple) >> Collect() == []
    assert iter(input) >> Unzip() >> Map(tuple) >> Collect() == expected

    assert [] >> Unzip(tuple) >> Collect() == []
    assert iter(input) >> Unzip(tuple) >> Collect() == expected


def test_CountValues():
    assert [] >> CountValues() == dict()
    assert 'abaacc' >> CountValues() == {'a': 3, 'b': 1, 'c': 2}

    assert [] >> CountValues(relative=True) == dict()
    assert [2] >> CountValues(relative=True) == {2: 1.0}
    assert 'aabaab' >> CountValues(relative=True) == {'a': 1.0, 'b': 0.5}

    data = [('a', 'X'), ('b', 'Y'), ('a', 'Y')]
    assert data >> CountValues(column=0) == {'a': 2, 'b': 1}
    assert data >> CountValues(column=1) == {'Y': 2, 'X': 1}


def test_Collect():
    assert [] >> Collect() == []
    assert range(5) >> Collect() == [0, 1, 2, 3, 4]
    assert [1, 2, 3, 2] >> Collect(set) == {1, 2, 3}
    assert [('one', 1), ('two', 2)] >> Collect(dict) == {'one': 1, 'two': 2}


def test_Join():
    assert [] >> Join() == ''
    assert range(5) >> Join() == '01234'
    assert range(5) >> Join(',') == '0,1,2,3,4'


def test_WriteCSV():
    filepath = 'tests/data/data_out.csv'
    data = [[1, 2], [3, 4]]

    with WriteCSV(filepath) as writer:
        data >> writer
    assert_equal_text(open(filepath).read(), '1,2\n3,4\n')

    with WriteCSV(filepath, cols=(1, 0)) as writer:
        data >> writer
    assert_equal_text(open(filepath).read(), '2,1\n4,3\n')

    with WriteCSV(filepath, cols=1) as writer:
        data >> writer
    assert_equal_text(open(filepath).read(), '2\n4\n')

    with WriteCSV(filepath, fmtfunc=lambda x: x + 1) as writer:
        data >> writer
    assert_equal_text(open(filepath).read(), '2,3\n4,5\n')

    with WriteCSV(filepath, skipheader=1) as writer:
        data >> writer
    assert_equal_text(open(filepath).read(), '3,4\n')

    with WriteCSV(filepath) as writer:
        [1, 2, 3] >> writer
    assert_equal_text(open(filepath).read(), '1\n2\n3\n')

    os.remove(filepath)


def test_WriteCSV_flush():
    filepath = 'tests/data/data_out.tsv'
    data = [[1, 2], [3, 4]]

    with WriteCSV(filepath, delimiter='\t', flush=True) as writer:
        data >> writer
        assert_equal_text(open(filepath).read(), '1\t2\n3\t4\n')

    os.remove(filepath)


def test_WriteCSV_tsv():
    filepath = 'tests/data/data_out.tsv'
    data = [[1, 2], [3, 4]]

    with WriteCSV(filepath, delimiter='\t') as writer:
        data >> writer
    assert_equal_text(open(filepath).read(), '1\t2\n3\t4\n')

    os.remove(filepath)



"""
.. module:: factory
   :synopsis: Unit tests for factory module
"""

from six.moves import range 
from nutsflow.base import Nut
from nutsflow.sink import Collect
from nutsflow.factory import (_arg_insert, _create_nut_wrapper,
                              _create_filter_wrapper, nut_processor, nut_sink,
                              nut_function, nut_source, nut_filter,
                              nut_filterfalse)


def test_arg_insert():
    args = (1, 2)
    assert _arg_insert(args, 'a') == ['a', 1, 2]
    assert _arg_insert(args, 'a', 1) == [1, 'a', 2]
    assert _arg_insert(args, 'a', 3) == [1, 2, 'a']
    assert _arg_insert(args, 'a', None) == [1, 2, 'a']


def test_create_nut_wrapper():
    func = lambda iterable, arg1, arg2: (iterable, arg1, arg2)
    wrapper = _create_nut_wrapper(Nut, func, 0)
    assert isinstance(wrapper(), Nut)
    assert ['a'] >> wrapper(2, 3) >> Collect() == [['a'], 2, 3]


def test_create_filter_wrapper():
    greaterThan = lambda x, t: x > t
    wrapper = _create_filter_wrapper(greaterThan)
    assert isinstance(wrapper(), Nut)
    assert [1, 2, 3, 4] >> wrapper(2) >> Collect() == [3, 4]

    wrapper = _create_filter_wrapper(greaterThan, True)
    assert [1, 2, 3, 4] >> wrapper(2) >> Collect() == [1, 2]

    not_empty = lambda x: x
    wrapper = _create_filter_wrapper(not_empty)
    assert [[1], [], [3], []] >> wrapper() >> Collect() == [[1], [3]]
    assert [1, 0, 3, 0] >> wrapper() >> Collect() == [1, 3]
    assert [1, [], [3], []] >> wrapper() >> Collect() == [1, [3]]


def test_nut_processor():
    @nut_processor
    def CloneN(iterable, n):
        for e in iterable:
            for _ in range(n):
                yield e

    assert [1, 2] >> CloneN(2) >> Collect() == [1, 1, 2, 2]


def test_nut_sink():
    @nut_sink
    def MyCollect(iterable, container):
        return container(iterable)

    assert [1, 2, 2] >> MyCollect(set) == {1, 2}


def test_nut_function():
    @nut_function
    def TimesN(x, n):
        return x * n

    assert [1, 2] >> TimesN(2) >> Collect() == [2, 4]


def test_nut_source():
    @nut_source
    def MyRange(start, end):
        return iter(range(start, end))

    assert MyRange(1, 4) >> Collect() == [1, 2, 3]


def test_nut_filter():
    @nut_filter
    def GreaterThan(x, threshold):
        return x > threshold

    assert [1, 2, 3, 4] >> GreaterThan(2) >> Collect() == [3, 4]


def test_nut_filterfalse():
    @nut_filterfalse
    def NotGreaterThan(x, threshold):
        return x > threshold

    assert [1, 2, 3, 4] >> NotGreaterThan(2) >> Collect() == [1, 2]
